@@ -72,7 +72,6 @@ result file::open(native_string const& f, mode m, creation_flags d)
 	}
 
 	SECURITY_ATTRIBUTES attr{};
-	attr.bInheritHandle = false;
 	attr.nLength = sizeof(SECURITY_ATTRIBUTES);
 
 	security_descriptor_builder sdb;
@@ -82,7 +81,7 @@ result file::open(native_string const& f, mode m, creation_flags d)
 			sdb.add(security_descriptor_builder::administrators);
 		}
 
-		auto sd = sdb.get_sd();
+		auto sd = sdb.get_sd(sdb_flags::none);
 		if (!sd) {
 			return {result::other};
 		}
