@@ -5,6 +5,8 @@
   #include <winsock2.h>
   #include <ws2tcpip.h>
   #include <mstcpip.h>
+
+  #include "windows/poller.hpp"
 #endif
 
 #include "libfilezilla/socket.hpp"
@@ -15,6 +17,7 @@
 #ifndef FZ_WINDOWS
   #include "libfilezilla/glue/unix.hpp"
   #include "libfilezilla/process.hpp"
+  #include "unix/poller.hpp"
 
   #define mutex mutex_override // Sadly on some platforms system headers include conflicting names
   #include <sys/types.h>
@@ -28,11 +31,7 @@
   #if !defined(MSG_NOSIGNAL) && !defined(SO_NOSIGPIPE)
 	#include <signal.h>
   #endif
-  #include <poll.h>
   #undef mutex
-  #if HAVE_EVENTFD
-	#include <sys/eventfd.h>
-  #endif
   #if HAVE_TCP_INFO
 	#include <atomic>
   #endif
@@ -257,10 +256,10 @@ int last_socket_error()
 }
 #endif
 
+#ifdef FZ_WINDOWS
 int set_nonblocking(socket::socket_t fd)
 {
 	// Set socket to non-blocking.
-#ifdef FZ_WINDOWS
 	unsigned long nonblock = 1;
 	int res = ioctlsocket(fd, FIONBIO, &nonblock);
 	if (!res) {
@@ -269,18 +268,8 @@ int set_nonblocking(socket::socket_t fd)
 	else {
 		return convert_msw_error_code(WSAGetLastError());
 	}
-#else
-	int flags = fcntl(fd, F_GETFL);
-	if (flags == -1) {
-		return errno;
-	}
-	int res = fcntl(fd, F_SETFL, flags | O_NONBLOCK);
-	if (res == -1) {
-		return errno;
-	}
-	return 0;
-#endif
 }
+#endif
 
 int do_set_flags(socket::socket_t fd, int flags, int flags_mask, duration const& keepalive_interval)
 {
@@ -373,6 +362,18 @@ private:
 #endif
 
 
+void close_socket_fds(std::vector<socket::socket_t> & fds_to_close)
+{
+	for (auto fd : fds_to_close) {
+#ifdef FZ_WINDOWS
+		closesocket(fd);
+#else
+		close(fd);
+#endif
+	}
+	fds_to_close.clear();
+}
+
 void close_socket_fd(socket::socket_t& fd)
 {
 	if (fd != -1) {
@@ -407,72 +408,15 @@ public:
 	~socket_thread()
 	{
 		thread_.join();
-		destroy_sync();
+		close_socket_fds(fds_to_close_);
 	}
 
 	int create_sync()
 	{
-#ifdef FZ_WINDOWS
-		if (sync_event_ == WSA_INVALID_EVENT) {
-			sync_event_ = WSACreateEvent();
-		}
-		if (sync_event_ == WSA_INVALID_EVENT) {
-			return 1;
-		}
-#elif defined HAVE_EVENTFD
-		if (event_fd_ == -1) {
-			event_fd_ = eventfd(0, EFD_CLOEXEC|EFD_NONBLOCK);
-			if (event_fd_ == -1) {
-				return errno;
-			}
-#ifndef HAVE_POLL
-			if (event_fd_ >= FD_SETSIZE) {
-				destroy_sync();
-				return EMFILE;
-			}
-#endif
-		}
-#else
-		if (pipe_[0] == -1) {
-			if (!create_pipe(pipe_)) {
-				return errno;
-			}
-
-#ifndef HAVE_POLL
-			if (pipe_[0] >= FD_SETSIZE) {
-				destroy_sync();
-				return EMFILE;
-			}
-#endif
-		}
-#endif
-		return 0;
+		return poller_.init();
 	}
 
 
-	void destroy_sync()
-	{
-#ifdef FZ_WINDOWS
-		if (sync_event_ != WSA_INVALID_EVENT) {
-			WSACloseEvent(sync_event_);
-		}
-#elif defined HAVE_EVENTFD
-		if (event_fd_ != -1) {
-			::close(event_fd_);
-			event_fd_ = -1;
-		}
-#else
-		if (pipe_[0] != -1) {
-			::close(pipe_[0]);
-			pipe_[0] = -1;
-		}
-		if (pipe_[1] != -1) {
-			::close(pipe_[1]);
-			pipe_[1] = -1;
-		}
-#endif
-	}
-
 	void set_socket(socket_base* pSocket)
 	{
 		scoped_lock l(mutex_);
@@ -511,7 +455,6 @@ public:
 	{
 		if (thread_) {
 			scoped_lock l(mutex_);
-			assert(threadwait_);
 			waiting_ = 0;
 			wakeup_thread(l);
 			return 0;
@@ -525,14 +468,13 @@ public:
 		thread_ = socket_->thread_pool_.spawn([this]() { entry(); });
 
 		if (!thread_) {
-			destroy_sync();
 			return EMFILE;
 		}
 
 		return 0;
 	}
 
-	// Cancels select or idle wait
+	// Cancels poll or idle wait
 	void wakeup_thread()
 	{
 		scoped_lock l(mutex_);
@@ -545,29 +487,7 @@ public:
 			return;
 		}
 
-		if (threadwait_) {
-			threadwait_ = false;
-			condition_.signal(l);
-			return;
-		}
-
-#ifdef FZ_WINDOWS
-		WSASetEvent(sync_event_);
-#elif defined(HAVE_EVENTFD)
-		uint64_t tmp = 1;
-
-		int ret;
-		do {
-			ret = write(event_fd_, &tmp, 8);
-		} while (ret == -1 && errno == EINTR);
-#else
-		char tmp = 0;
-
-		int ret;
-		do {
-			ret = write(pipe_[1], &tmp, 1);
-		} while (ret == -1 && errno == EINTR);
-#endif
+		poller_.interrupt(l);
 	}
 
 protected:
@@ -593,14 +513,6 @@ protected:
 #endif
 		}
 
-#if !defined(FZ_WINDOWS) && !defined(HAVE_POLL)
-		if (fd >= FD_SETSIZE) {
-			close(fd);
-			errno = EMFILE;
-			return -1;
-		}
-#endif
-
 		if (fd != -1) {
 #if defined(SO_NOSIGPIPE) && !defined(MSG_NOSIGNAL)
 			// We do not want SIGPIPE if writing to socket.
@@ -811,51 +723,43 @@ protected:
 
 		for (;;) {
 #ifdef FZ_WINDOWS
-			int wait_events{};
+			pollinfo info;
+			info.fd_ = socket_->fd_;
+			
 			if (waiting_ & WAIT_CONNECT) {
-				wait_events |= FD_CONNECT;
+				info.events_ |= FD_CONNECT;
 			}
 			if (waiting_ & WAIT_READ) {
-				wait_events |= FD_READ | FD_CLOSE;
+				info.events_ |= FD_READ | FD_CLOSE;
 			}
 			if (waiting_ & WAIT_WRITE) {
-				wait_events |= FD_WRITE;
+				info.events_ |= FD_WRITE;
 			}
 			if (waiting_ & WAIT_ACCEPT) {
-				wait_events |= FD_ACCEPT;
+				info.events_ |= FD_ACCEPT;
 			}
-			WSAEventSelect(socket_->fd_, sync_event_, wait_events);
-			l.unlock();
-			WSAWaitForMultipleEvents(1, &sync_event_, false, WSA_INFINITE, false);
+			bool res = poller_.wait(&info, 1, l);
 
-			l.lock();
-			if (should_quit()) {
-				return false;
-			}
-
-			WSANETWORKEVENTS events;
-			int res = WSAEnumNetworkEvents(socket_->fd_, sync_event_, &events);
-			if (res) {
-				res = last_socket_error();
+			if (!res || should_quit() || socket_->fd_ == -1) {
 				return false;
 			}
 
 			if (waiting_ & WAIT_CONNECT) {
-				if (events.lNetworkEvents & FD_CONNECT) {
+				if (info.result_.lNetworkEvents & FD_CONNECT) {
 					triggered_ |= WAIT_CONNECT;
-					triggered_errors_[0] = convert_msw_error_code(events.iErrorCode[FD_CONNECT_BIT]);
+					triggered_errors_[0] = convert_msw_error_code(info.result_.iErrorCode[FD_CONNECT_BIT]);
 					waiting_ &= ~WAIT_CONNECT;
 				}
 			}
 			if (waiting_ & WAIT_READ) {
-				if (events.lNetworkEvents & FD_READ) {
+				if (info.result_.lNetworkEvents & FD_READ) {
 					triggered_ |= WAIT_READ;
-					triggered_errors_[1] = convert_msw_error_code(events.iErrorCode[FD_READ_BIT]);
+					triggered_errors_[1] = convert_msw_error_code(info.result_.iErrorCode[FD_READ_BIT]);
 					waiting_ &= ~WAIT_READ;
 				}
-				if (events.lNetworkEvents & FD_CLOSE) {
+				if (info.result_.lNetworkEvents & FD_CLOSE) {
 					triggered_ |= WAIT_READ;
-					int err = convert_msw_error_code(events.iErrorCode[FD_CLOSE_BIT]);
+					int err = convert_msw_error_code(info.result_.iErrorCode[FD_CLOSE_BIT]);
 					if (err) {
 						triggered_errors_[1] = err;
 					}
@@ -863,16 +767,16 @@ protected:
 				}
 			}
 			if (waiting_ & WAIT_WRITE) {
-				if (events.lNetworkEvents & FD_WRITE) {
+				if (info.result_.lNetworkEvents & FD_WRITE) {
 					triggered_ |= WAIT_WRITE;
-					triggered_errors_[2] = convert_msw_error_code(events.iErrorCode[FD_WRITE_BIT]);
+					triggered_errors_[2] = convert_msw_error_code(info.result_.iErrorCode[FD_WRITE_BIT]);
 					waiting_ &= ~WAIT_WRITE;
 				}
 			}
 			if (waiting_ & WAIT_ACCEPT) {
-				if (events.lNetworkEvents & FD_ACCEPT) {
+				if (info.result_.lNetworkEvents & FD_ACCEPT) {
 					triggered_ |= WAIT_ACCEPT;
-					triggered_errors_[3] = convert_msw_error_code(events.iErrorCode[FD_ACCEPT_BIT]);
+					triggered_errors_[3] = convert_msw_error_code(info.result_.iErrorCode[FD_ACCEPT_BIT]);
 					waiting_ &= ~WAIT_ACCEPT;
 				}
 			}
@@ -880,58 +784,24 @@ protected:
 			if (triggered_ || !waiting_) {
 				return true;
 			}
-#elif defined(HAVE_POLL)
-			pollfd fds[2]{};
-#ifdef HAVE_EVENTFD
-			fds[0].fd = event_fd_;
 #else
-			fds[0].fd = pipe_[0];
-#endif
-			fds[0].events = POLLIN;
-			fds[1].fd = socket_->fd_;
-			fds[1].events = 0;
+			pollfd fds[2]{};
+			fds[0].fd = socket_->fd_;
+			fds[0].events = 0;
 
 			if (waiting_ & (WAIT_READ|WAIT_ACCEPT)) {
-				fds[1].events |= POLLIN;
+				fds[0].events |= POLLIN;
 			}
 			if (waiting_ & (WAIT_WRITE | WAIT_CONNECT)) {
-				fds[1].events |= POLLOUT;
+				fds[0].events |= POLLOUT;
 			}
+			bool res = poller_.wait(fds, 1, l);
 
-			l.unlock();
-
-			int res = poll(fds, 2, -1);
-
-			l.lock();
-
-			if (res > 0 && fds[0].revents) {
-				char buffer[100];
-#ifdef HAVE_EVENTFD
-				int damn_spurious_warning = read(event_fd_, buffer, 8);
-#else
-				int damn_spurious_warning = read(pipe_[0], buffer, 100);
-#endif
-				(void)damn_spurious_warning; // We do not care about return value and this is definitely correct!
-			}
-
-			if (quit_ || !socket_ || socket_->fd_ == -1) {
+			if (!res || should_quit() || socket_->fd_ == -1) {
 				return false;
 			}
 
-			if (!res) {
-				continue;
-			}
-			if (res == -1) {
-				res = errno;
-
-				if (res == EINTR) {
-					continue;
-				}
-
-				return false;
-			}
-
-			int const revents = fds[1].revents;
+			int const revents = fds[0].revents;
 			if (waiting_ & WAIT_CONNECT) {
 				if (revents & (POLLOUT|POLLERR|POLLHUP)) {
 					int error;
@@ -966,103 +836,6 @@ protected:
 				}
 			}
 
-			if (triggered_ || !waiting_) {
-				return true;
-			}
-#else
-			fd_set readfds;
-			fd_set writefds;
-			FD_ZERO(&readfds);
-			FD_ZERO(&writefds);
-
-#ifdef HAVE_EVENTFD
-			FD_SET(event_fd_, &readfds);
-#else
-			FD_SET(pipe_[0], &readfds);
-#endif
-
-			if (waiting_ & (WAIT_READ | WAIT_ACCEPT)) {
-				FD_SET(socket_->fd_, &readfds);
-			}
-			if (waiting_ & (WAIT_WRITE | WAIT_CONNECT)) {
-				FD_SET(socket_->fd_, &writefds);
-			}
-
-#ifdef HAVE_EVENTFD
-			int maxfd = std::max(event_fd_, socket_->fd_) + 1;
-#else
-			int maxfd = std::max(pipe_[0], socket_->fd_) + 1;
-#endif
-
-			l.unlock();
-
-			int res = select(maxfd, &readfds, &writefds, nullptr, nullptr);
-
-			l.lock();
-
-#ifdef HAVE_EVENTFD
-			if (res > 0 && FD_ISSET(event_fd_, &readfds)) {
-				char buffer[8];
-				int damn_spurious_warning = read(event_fd_, buffer, 8);
-#else
-			if (res > 0 && FD_ISSET(pipe_[0], &readfds)) {
-				char buffer[100];
-				int damn_spurious_warning = read(pipe_[0], buffer, 100);
-#endif
-				(void)damn_spurious_warning; // We do not care about return value and this is definitely correct!
-			}
-
-			if (quit_ || !socket_ || socket_->fd_ == -1) {
-				return false;
-			}
-
-			if (!res) {
-				continue;
-			}
-			if (res == -1) {
-				res = errno;
-
-				if (res == EINTR) {
-					continue;
-				}
-
-				return false;
-			}
-
-			if (waiting_ & WAIT_CONNECT) {
-				if (FD_ISSET(socket_->fd_, &writefds)) {
-					int error;
-					socklen_t len = sizeof(error);
-					int getsockopt_res = getsockopt(socket_->fd_, SOL_SOCKET, SO_ERROR, &error, &len);
-					if (getsockopt_res) {
-						error = errno;
-					}
-					triggered_ |= WAIT_CONNECT;
-					triggered_errors_[0] = error;
-					waiting_ &= ~WAIT_CONNECT;
-				}
-			}
-			else if (waiting_ & WAIT_ACCEPT) {
-				if (FD_ISSET(socket_->fd_, &readfds)) {
-					triggered_ |= WAIT_ACCEPT;
-					waiting_ &= ~WAIT_ACCEPT;
-				}
-			}
-			else {
-				if (waiting_ & WAIT_READ) {
-					if (FD_ISSET(socket_->fd_, &readfds)) {
-						triggered_ |= WAIT_READ;
-						waiting_ &= ~WAIT_READ;
-					}
-				}
-				if (waiting_ & WAIT_WRITE) {
-					if (FD_ISSET(socket_->fd_, &writefds)) {
-						triggered_ |= WAIT_WRITE;
-						waiting_ &= ~WAIT_WRITE;
-					}
-				}
-			}
-
 			if (triggered_ || !waiting_) {
 				return true;
 			}
@@ -1097,8 +870,8 @@ protected:
 			return false;
 		}
 		while (!socket_ || (!waiting_ && host_.empty())) {
-			threadwait_ = true;
-			condition_.wait(l);
+
+			poller_.wait(l);
 
 			if (quit_) {
 				return false;
@@ -1149,6 +922,7 @@ protected:
 					send_events();
 				}
 			}
+			close_socket_fds(fds_to_close_);
 		}
 
 		if (thread_) {
@@ -1168,19 +942,12 @@ protected:
 	std::string bind_;
 
 	mutex mutex_;
-	condition condition_;
 
 	async_task thread_;
 
-#ifdef FZ_WINDOWS
-	// We wait on this using WSAWaitForMultipleEvents
-	WSAEVENT sync_event_{WSA_INVALID_EVENT};
-#elif defined(HAVE_EVENTFD)
-	int event_fd_{-1};
-#else
-	// A pipe is used to unblock select
-	int pipe_[2]{-1, -1};
-#endif
+	std::vector<socket::socket_t> fds_to_close_;
+
+	poller poller_;
 
 	// The socket events we are waiting for
 	int waiting_{};
@@ -1190,9 +957,6 @@ protected:
 	int triggered_errors_[WAIT_EVENTCOUNT];
 
 	bool quit_{};
-
-	// Thread waits for instructions
-	bool threadwait_{};
 };
 
 socket_base::socket_base(thread_pool& pool, event_handler* evt_handler, socket_event_source* ev_source)
@@ -1245,15 +1009,16 @@ int socket_base::close()
 	}
 
 	scoped_lock l(socket_thread_->mutex_);
-	socket_t fd = fd_;
-	fd_ = -1;
+	if (fd_ != -1) {
+		socket_thread_->fds_to_close_.emplace_back(fd_);
+		fd_ = -1;
+	}
 
 	socket_thread_->host_.clear();
 	socket_thread_->port_.clear();
 
 	socket_thread_->wakeup_thread(l);
 
-	close_socket_fd(fd);
 	if (dynamic_cast<socket*>(this)) {
 		static_cast<socket*>(this)->state_ = socket_state::closed;
 	}
@@ -1603,14 +1368,6 @@ socket_descriptor listen_socket::fast_accept(int &error)
 		if (fd == -1) {
 			error = errno;
 		}
-
-#if !defined(FZ_WINDOWS) && !defined(HAVE_POLL)
-		if (fd >= FD_SETSIZE) {
-			::close(fd);
-			error = EMFILE;
-			return socket_descriptor();
-		}
-#endif
 	}
 
 	if (fd != -1) {
